import json
import sys
import time
import uuid
from dataclasses import dataclass
from traceback import TracebackException
from typing import Optional, Any, Union

from onestep._utils import catch_error


class MessageTracebackException(TracebackException):
    def __init__(self, exc_type, exc_value, exc_traceback, **kwargs):
        super().__init__(exc_type, exc_value, exc_traceback, **kwargs)
        self.exc_value = exc_value


@dataclass
class Extra:
    task_id: str = None
    publish_time: float = None
    failure_count: int = 0

    def __post_init__(self):
        self.task_id = self.task_id or str(uuid.uuid4())
        self.publish_time = self.publish_time or round(time.time(), 3)

    def to_dict(self):
        return {
            'task_id': self.task_id,
            'publish_time': self.publish_time,
            'failure_count': self.failure_count,
        }

    def __str__(self):
        return str(self.to_dict())


class Message:

    def __init__(
            self,
            body: Optional[Union[dict, Any]] = None,
            extra: Optional[Union[dict, Extra]] = None,
            message: Optional[Any] = None,
            broker=None
    ):
        """ Message

        :param body: 解析后的消息体
        :param extra: 额外信息
        :param message: 原始消息体
        :param broker: 当前消息所属的 broker
        """
        self.body = body
        self.extra = self._set_extra(extra)
        self.message = message

        self.broker = broker
        self._exception = None

    @staticmethod
    def _set_extra(extra):
        if isinstance(extra, Extra):
            return extra
        elif isinstance(extra, dict):
            return Extra(**extra)
        else:
            return Extra()

    def set_exception(self):
        """设置异常信息，会自动获取"""
        self._exception = MessageTracebackException(*sys.exc_info())
        self.failure_count = self.failure_count + 1

    @property
    def exception(self) -> Optional[MessageTracebackException]:
        return self._exception

    @property
    def fail(self):
        return self.exception is not None

    @property
    def failure_count(self):
        return self.extra.failure_count

    @failure_count.setter
    def failure_count(self, value):
        self.extra.failure_count = value

    def replace(self, **kwargs):
        """替换当前message的属性"""
        for key, value in kwargs.items():
            if not hasattr(self, key):
                continue
            if key == 'extra':
                value = self._set_extra(value)
            setattr(self, key, value)
        return self

    def to_dict(self, include_exception=False) -> dict:
        data = {'body': self.body, 'extra': self.extra.to_dict()}
        if include_exception and self.exception:
            data['exception'] = "".join(self.exception.format(chain=True))  # noqa

        return data

    def to_json(self, include_exception=False) -> str:
        return json.dumps(self.to_dict(include_exception))

    @catch_error()
    def confirm(self):
        """确认消息"""
        self.broker.confirm(self)

    @catch_error()
    def reject(self):
        """拒绝消息"""
        self.broker.reject(self)

    @catch_error()
    def requeue(self, is_source=False):
        """
        重发消息：先拒绝 再 重入
        
        :param is_source: 是否是源消息，True: 使用消息的最新数据重入当前队列，False: 使用消息的最新数据重入当前队列
        """
        self.broker.requeue(self, is_source=is_source)

    def __getattr__(self, item):
        return None

    def __delattr__(self, item):
        if hasattr(self, item):
            setattr(self, item, None)

    def __str__(self):
        return str(self.to_dict())

    def __repr__(self):
        return f"<{self.__class__.__name__} {self.body}>"

    @classmethod
    def from_broker(cls, broker_message: Any):
        return cls(body=broker_message, extra=None, message=broker_message, broker=None)



import argparse
import importlib
import logging
import sys
from onestep import step, __version__

LOGFORMAT = "[%(asctime)s] [%(name)s] [%(levelname)s] %(message)s"


def setup_logging():
    logging.basicConfig(level=logging.DEBUG, format=LOGFORMAT, stream=sys.stdout)

    # exclude amqpstorm logs
    logging.getLogger("amqpstorm").setLevel(logging.CRITICAL)
    return logging.getLogger("onestep")


logger = setup_logging()


def parse_args():
    parser = argparse.ArgumentParser(
        description='run onestep'
    )
    parser.add_argument(
        "step",
        help="the run step",
    )
    parser.add_argument(
        "--group", "-G", default=None,
        help="the run group",
        type=str
    )
    parser.add_argument(
        "--print",
        action="store_true",
        help="enable printing")
    parser.add_argument(
        "--path", "-P", default=".", nargs="*", type=str,
        help="the step import path (default: current running directory)"
    )
    return parser.parse_args()


def main():
    args = parse_args()
    for path in args.path:
        sys.path.insert(0, path)
    logger.info(f"OneStep {__version__} is start up.")
    try:
        importlib.import_module(args.step)
        step.start(group=args.group, block=True, print_jobs=args.print)
    except KeyboardInterrupt:
        step.shutdown()


if __name__ == '__main__':
    sys.exit(main())



import logging
import json
from queue import Queue, Full as FullException
from typing import Any

from .base import BaseBroker, BaseConsumer

from ..message import Message

logger = logging.getLogger(__name__)


class MemoryMessage(Message):

    @classmethod
    def from_broker(cls, broker_message: Any):
        if isinstance(broker_message, (str, bytes, bytearray)):
            try:
                message = json.loads(broker_message)
            except json.JSONDecodeError:
                message = {"body": broker_message}
        else:
            message = broker_message
        if not isinstance(message, dict):
            message = {"body": message}
        if "body" not in message:
            # 来自 外部的消息 可能没有 body, 故直接认为都是 message.body
            message = {"body": message}
        return cls(body=message.get("body"), extra=message.get("extra"), message=broker_message)


class MemoryBroker(BaseBroker):
    message_cls = MemoryMessage

    def __init__(self, maxsize=0, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.queue = Queue(maxsize)

    def publish(self, message: Any):
        try:
            self.queue.put_nowait(message)
        except FullException:
            logger.warning("CronBroker queue is full, skip this task, "
                           "you can increase maxsize with `maxsize` argument")

    def consume(self, *args, **kwargs):
        return MemoryConsumer(self, *args, **kwargs)

    def confirm(self, message: Message):
        """确认消息"""
        pass

    def reject(self, message: Message):
        """拒绝消息"""
        pass

    def requeue(self, message: Message, is_source=False):
        """重发消息：先拒绝 再 重入"""
        if is_source:
            self.publish(message.message)
        else:
            self.send(message)

    def __repr__(self):
        return f"<{self.__class__.__name__} {self.name}>"

    def __str__(self):
        return self.name


class MemoryConsumer(BaseConsumer):
    ...



"""
将指定的函数放入线程中运行
"""
from concurrent.futures import ThreadPoolExecutor
from typing import Dict, Iterable

import logging
import threading
from asyncio import iscoroutinefunction
from inspect import isasyncgenfunction

from asgiref.sync import async_to_sync

from .message import Message
from .retry import RetryStatus
from .broker import BaseBroker
from . import exception
from .signal import message_received, message_consumed, message_error, message_drop, message_requeue

logger = logging.getLogger(__name__)


class BaseWorker:
    broker_exit: Dict[BaseBroker, bool] = {}
    broker_exit_lock = threading.Lock()

    def __init__(self, onestep, broker: BaseBroker, *args, **kwargs):
        self.instance = onestep
        self.retry = self.instance.retry
        self.error_callback = self.instance.error_callback
        self.broker = broker
        self.args = args
        self.kwargs = kwargs
        self._shutdown = False

    @property
    def instance_name(self):
        return self.instance.fn.__name__

    def start(self):
        """启动 Worker"""
        raise NotImplementedError

    def run(self):
        """执行 Worker 的逻辑"""
        raise NotImplementedError

    def shutdown(self):
        """关闭 Worker"""
        raise NotImplementedError

    def receive_messages(self) -> Iterable[Message]:
        """ 从broker中获取消息 """
        for result in self.broker.consume():
            if self._shutdown:
                break
            if result is None:
                continue
            messages = result if isinstance(result, Iterable) else [result]
            yield from messages
            # when broker is once, it will shut down after receive a message
            if self.broker.once:
                self.shutdown()

    def _run_real_instance(self, message: Message) -> None:
        """ 执行实例的逻辑 """
        if iscoroutinefunction(self.instance.fn) or isasyncgenfunction(self.instance.fn):
            async_to_sync(self.instance)(message, *self.args, **self.kwargs)
        else:
            self.instance(message, *self.args, **self.kwargs)

    def handle_message(self, message: Message):
        """ 处理消息 """
        message.broker = message.broker or self.broker
        logger.debug(f"{self.instance.name} receive message<{message}> from {self.broker!r}")
        message_received.send(self, message=message)
        try:
            self.instance.before_emit("consume", message=message)

            self._run_real_instance(message)
            message_consumed.send(self, message=message)
            message.confirm()

            self.instance.after_emit("consume", message=message)
        except (exception.DropMessage, exception.RejectMessage) as e:
            message_drop.send(self, message=message, reason=e)
            logger.warning(f"{self.instance.name} dropped <{type(e).__name__}: {str(e)}>")
            message.reject()
        except exception.RequeueMessage as e:
            message_requeue.send(self, message=message, reason=e)
            logger.warning(f"{self.instance.name} requeue <{type(e).__name__}: {str(e)}>")
            message.requeue(is_source=True)
        except Exception as e:
            message_error.send(self, message=message, error=e)
            if self.instance.state.debug:
                logger.exception(f"{self.instance.name} run error <{type(e).__name__}: {str(e)}>")
            else:
                logger.error(f"{self.instance.name} run error <{type(e).__name__}: {str(e)}>")
            message.set_exception()

            retry_status = self.retry(message)
            if retry_status is RetryStatus.END_WITH_CALLBACK:
                if self.error_callback:
                    self.error_callback(message)
                message.reject()
            elif retry_status is RetryStatus.END_IGNORE_CALLBACK:
                # 由于是队列内重试，不会触发错误回调
                message.requeue()
        finally:
            # When message is triggered by cancel_consume, it will be shutdown
            if self.broker.cancel_consume and self.broker.cancel_consume(message):
                self.shutdown()

    def __repr__(self):
        return f"<{self.__class__.__name__} {self.instance.name}>"


class ThreadWorker(BaseWorker):

    def __init__(self, onestep, broker: BaseBroker, *args, **kwargs):
        """
        线程执行包装过的`onestep`函数
        :param onestep: OneStep实例
        :param broker: 监听的from broker
        """
        super().__init__(onestep, broker, *args, **kwargs)
        self.thread = None

    def start(self):
        """启动单线程 Worker"""
        self.thread = threading.Thread(target=self.run, daemon=True)
        self.thread.start()

    def run(self):
        """线程执行包装过的`onestep`函数

        `fn`为`onestep`函数，执行会调用`onestep`的`__call__`方法
        :return:
        """

        while not self._shutdown:
            with ThreadWorker.broker_exit_lock:
                if ThreadWorker.broker_exit.get(self.broker, False):
                    self.shutdown()
                    break
            for message in self.receive_messages():
                self.handle_message(message)

    def shutdown(self):
        ThreadWorker.broker_exit[self.broker] = True
        self.broker.shutdown()
        self._shutdown = True


class ThreadPoolWorker(BaseWorker):

    def __init__(self, onestep, broker: BaseBroker, workers=None, *args, **kwargs):
        super().__init__(onestep, broker, *args, **kwargs)
        self.executor = ThreadPoolExecutor(max_workers=workers)

    def start(self):
        """启动线程池 Worker"""
        self.executor.submit(self.run)

    def run(self):
        """线程执行包装过的`onestep`函数

        `fn`为`onestep`函数，执行会调用`onestep`的`__call__`方法
        :return:
        """

        while not self._shutdown:
            with ThreadPoolWorker.broker_exit_lock:
                if ThreadPoolWorker.broker_exit.get(self.broker, False):
                    self.shutdown()
                    break
            for message in self.receive_messages():
                self.handle_message(message)

    def shutdown(self):
        """关闭线程池 Worker"""
        ThreadPoolWorker.broker_exit[self.broker] = True
        self._shutdown = True
        self.executor.shutdown()



import hashlib
import json

from . import BaseMiddleware
from ..exception import DropMessage


class hash_func:
    """ Default message to hash function """

    def __call__(self, body):
        data = json.dumps(body) if isinstance(body, dict) else str(body)
        return hashlib.sha1(data.encode("utf-8")).hexdigest()


class UniqueMiddleware(BaseMiddleware):
    default_hash_func = hash_func()

    def before_consume(self, step, message, *args, **kwargs):
        message_hash = self.default_hash_func(message.body)
        if self.has_seen(message_hash):
            raise DropMessage(f"Message<{message}> has been seen before")

    def after_consume(self, step, message, *args, **kwargs):
        if message.fail or message.body is None:
            return
        message_hash = self.default_hash_func(message.body)
        self.mark_seen(message_hash)

    def has_seen(self, message):
        raise NotImplementedError

    def mark_seen(self, message):
        raise NotImplementedError


class MemoryUniqueMiddleware(UniqueMiddleware):
    """基于内存的去重中间件"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.seen = set()

    def has_seen(self, hash_value):
        return hash_value in self.seen

    def mark_seen(self, hash_value):
        self.seen.add(hash_value)



import json
import threading
import uuid
from queue import Queue
from typing import Optional, Dict, Any

try:
    from use_redis import useRedisStreamStore, RedisStreamMessage
except ImportError:
    ...

from ..base import BaseBroker, BaseConsumer, Message


class _RedisStreamMessage(Message):
    @classmethod
    def from_broker(cls, broker_message: "RedisStreamMessage"):
        if "_message" in broker_message.body:
            # 来自 RedisStreamBroker.send 的消息，message.body 默认是存于 _message 字段中
            try:
                message = json.loads(broker_message.body.get("_message"))  # 已转换的 message
            except (json.JSONDecodeError, TypeError):
                message = {"body": broker_message.body.get("_message")}  # 未转换的 message
        else:
            # 来自 外部的消息 直接认为都是 message.body
            message = {"body": broker_message.body}

        yield cls(body=message.get("body"), extra=message.get("extra"), message=broker_message)


class RedisStreamBroker(BaseBroker):
    """ Redis Stream Broker """
    message_cls = _RedisStreamMessage

    def __init__(
            self,
            stream: str,
            group: str = "onestep",
            params: Optional[Dict] = None,
            prefetch: Optional[int] = 1,
            stream_max_entries: int = 0,
            redeliver_timeout: int = 60000,
            claim_interval: int = 1800000,
            *args,
            **kwargs):
        super().__init__(*args, **kwargs)
        self.stream = stream
        self.group = group
        self.prefetch = prefetch
        self.queue = Queue()

        self.threads = []

        self.client = useRedisStreamStore(
            stream=stream,
            group=group,
            stream_max_entries=stream_max_entries,
            redeliver_timeout=redeliver_timeout,
            claim_interval=claim_interval,
            **(params or {})
        )

    def _consume(self, *args, **kwargs):
        def callback(message):
            self.queue.put(message)

        prefetch = kwargs.pop("prefetch", self.prefetch)
        self.client.start_consuming(consumer=uuid.uuid4().hex, callback=callback, prefetch=prefetch, **kwargs)

    def consume(self, *args, **kwargs):
        daemon = kwargs.pop('daemon', True)
        thread = threading.Thread(target=self._consume, *args, **kwargs)
        thread.daemon = daemon
        thread.start()
        self.threads.append(thread)
        return RedisStreamConsumer(self)

    def send(self, message: Any):
        """对消息进行预处理，然后再发送"""
        if not isinstance(message, Message):
            message = self.message_cls(body=message)

        self.client.send({"_message": message.to_json()})

    publish = send

    def confirm(self, message: Message):
        self.client.ack(message.message)

    def reject(self, message: Message):
        self.client.reject(message.message)

    def requeue(self, message: Message, is_source=False):
        """
         重发消息：先拒绝 再 重入

         :param message: 消息
         :param is_source: 是否是原始消息消息，True: 使用原始消息重入当前队列，False: 使用消息的最新数据重入当前队列
         """
        self.reject(message)

        if is_source:
            self.client.send(message.message.body)
        else:
            self.send(message)


class RedisStreamConsumer(BaseConsumer):
    ...



class DeprecationMeta(type):

    def __new__(cls, name, bases, attrs):
        if "before_receive" in attrs or "after_receive" in attrs:
            raise DeprecationWarning(
                f"`{name}` "
                "The before_receive and after_receive methods are deprecated, "
                "please use before_consume and after_consume instead"
            )
        return super().__new__(cls, name, bases, attrs)


class BaseMiddleware(metaclass=DeprecationMeta):

    def before_send(self, step, message, *args, **kwargs):
        """消息发送之前"""
        pass

    def after_send(self, step, message, *args, **kwargs):
        """消息发送之后"""
        pass

    def before_consume(self, step, message, *args, **kwargs):
        """消费消息之前"""
        pass

    def after_consume(self, step, message, *args, **kwargs):
        """消费消息之后"""
        pass

    def __repr__(self):
        return f"<{self.__class__.__name__}>"



import json
import threading
from queue import Queue
from typing import Any

try:
    from use_redis import useRedis
except ImportError:
    ...

from ..base import BaseBroker, BaseConsumer, Message


class _RedisPubSubMessage(Message):

    @classmethod
    def from_broker(cls, broker_message: Any):
        if "channel" in broker_message:
            try:
                message = json.loads(broker_message.get("data"))  # 已转换的 message
            except (json.JSONDecodeError, TypeError):
                message = {"body": broker_message.get("data")}  # 未转换的 message
        else:
            # 来自 外部的消息 直接认为都是 message.body
            message = {"body": broker_message.body}

        yield cls(body=message.get("body"), extra=message.get("extra"), message=broker_message)


class RedisPubSubBroker(BaseBroker):
    """ Redis PubSub Broker """
    message_cls = _RedisPubSubMessage

    def __init__(self, channel: str, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.channel = channel
        self.queue = Queue()

        self.threads = []

        self.client = useRedis(**kwargs).connection

    def _consume(self):
        def callback(message: dict):
            if message.get('type') != 'message':
                return
            self.queue.put(message)

        ps = self.client.pubsub()
        ps.subscribe(self.channel)
        for message in ps.listen():
            callback(message)

    def consume(self, *args, **kwargs):
        daemon = kwargs.pop('daemon', True)
        thread = threading.Thread(target=self._consume, *args, **kwargs)
        thread.daemon = daemon
        thread.start()
        self.threads.append(thread)
        return RedisPubSubConsumer(self)

    def send(self, message: Any):
        """Publish message to the Redis channel"""
        if not isinstance(message, Message):
            message = self.message_cls(body=message)

        self.client.publish(self.channel, message.to_json())

    publish = send

    def confirm(self, message: Message):
        pass

    def reject(self, message: Message):
        pass

    def requeue(self, message: Message, is_source=False):
        """
         重发消息：先拒绝 再 重入

         :param message: 消息
         :param is_source: 是否是原始消息消息，True: 使用原始消息重入当前队列，False: 使用消息的最新数据重入当前队列
         """
        self.reject(message)

        if is_source:
            self.client.publish(self.channel, message.message['data'])
        else:
            self.send(message)


class RedisPubSubConsumer(BaseConsumer):
    ...



from .base import (
    BaseBroker, BaseConsumer
)
from .memory import MemoryBroker, MemoryConsumer
from .webhook import WebHookBroker
from .rabbitmq import RabbitMQBroker
from .redis import RedisStreamBroker, RedisPubSubBroker
from .cron import CronBroker



import functools
import collections
import inspect
import logging

from inspect import isgenerator, iscoroutinefunction, isasyncgenfunction, isasyncgen
from itertools import groupby
from typing import Optional, List, Dict, Any, Callable, Union, Type

from .broker.base import BaseBroker
from .exception import StopMiddleware
from .message import Message
from .retry import TimesRetry
from .signal import message_sent, started, stopped
from .state import State
from .worker import ThreadWorker, BaseWorker

logger = logging.getLogger(__name__)

MAX_WORKERS = 20
DEFAULT_WORKERS = 1
DEFAULT_WORKER_CLASS = ThreadWorker


class BaseOneStep:
    consumers: Dict[str, List[BaseWorker]] = collections.defaultdict(list)
    state = State()  # 全局状态

    def __init__(self, fn,
                 group: str = "OneStep",
                 name: str = None,
                 from_broker: Union[BaseBroker, List[BaseBroker], None] = None,
                 to_broker: Union[BaseBroker, List[BaseBroker], None] = None,
                 workers: Optional[int] = None,
                 worker_class: Optional[Type[BaseWorker]] = None,
                 middlewares: Optional[List[Any]] = None,
                 retry: Union[Callable, object] = TimesRetry(),
                 error_callback: Optional[Union[Callable, object]] = None):
        self.group = group
        self.fn = fn
        self.name = name or fn.__name__
        self.workers = workers or DEFAULT_WORKERS
        self.worker_class = worker_class or DEFAULT_WORKER_CLASS
        if self.workers > MAX_WORKERS:
            logger.warning(f"workers[{self.workers}] litter than {MAX_WORKERS}")
            self.workers = MAX_WORKERS
        self.middlewares = middlewares or []

        self.from_brokers = self._init_broker(from_broker)
        self.to_brokers = self._init_broker(to_broker)
        self.retry = retry
        self.error_callback = error_callback

        for broker in self.from_brokers:
            self._add_consumer(broker)

    @staticmethod
    def _init_broker(broker: Union[BaseBroker, List[BaseBroker], None] = None):
        if not broker:
            return []

        if isinstance(broker, BaseBroker):
            return [broker]
        if isinstance(broker, (list, tuple)):
            return list(broker)
        raise TypeError(
            f"broker must be BaseBroker or list or tuple, not {type(broker)}")

    def _add_consumer(self, broker):
        """ 添加来源消费者 """
        worker_class_params = inspect.signature(self.worker_class.__init__).parameters
        if "workers" in worker_class_params:
            self.consumers[self.group].append(
                self.worker_class(onestep=self, broker=broker, workers=self.workers)
            )
        else:
            for _ in range(self.workers):
                self.consumers[self.group].append(
                    self.worker_class(onestep=self, broker=broker)
                )

    @classmethod
    def _find_consumers(cls, group: Optional[str] = None):
        """按组查找消费者"""
        if group is None:
            consumers = [c for v in cls.consumers.values() for c in v]
        else:
            consumers = cls.consumers[group]
        return consumers

    @classmethod
    def print_jobs(cls, group):
        print("Jobs:")
        prints = []
        _consumers = cls._find_consumers(group)
        group_instance = groupby(_consumers, key=lambda x: x.instance)
        for instance, _ in group_instance:
            prints.append([instance.name, instance.group, instance.workers, str(instance.from_brokers)])
        print("{:<15} {:<10} {:<10} {:<20}".format("Job", "Group", "Workers", "From Brokers"))
        for v in prints:
            print("{:<15} {:<10} {:<10} {:<20}".format(*v))

    @classmethod
    def start(cls, group: Optional[str] = None, print_jobs: bool = False):
        logger.debug(f"start group [{group or 'all'}]")
        _consumers = cls._find_consumers(group)
        if not _consumers:
            logger.debug(f"no consumer found in group [{group or 'all'}]")
            return
        if print_jobs:
            cls.print_jobs(group)
        for consumer in _consumers:
            consumer.start()
            logger.debug(f"started: {consumer=}")

    @classmethod
    def shutdown(cls, group: Optional[str] = None):
        logger.debug(f"stop group [{group or 'all'}]")
        for consumer in cls._find_consumers(group):
            consumer.shutdown()
            logger.debug(f"stopped: {consumer=}")

    def wraps(self, func):
        @functools.wraps(func)
        def wrapped_f(*args, **kwargs):
            return self(*args, **kwargs)  # noqa

        return wrapped_f

    def send(self, result, broker=None):
        """将返回的内容交给broker发送"""
        brokers = self._init_broker(broker) or self.to_brokers
        # 如果是Message类型，就不再封装
        message = result if isinstance(result, Message) else Message(body=result)
        self.before_emit("send", message=message)

        if result and not brokers:
            logger.debug("send(result): broker is empty")
            return

        if not result and brokers:
            logger.debug("send(result): body is empty")
            return

        for broker in brokers:
            message_sent.send(self, message=message, broker=broker)
            broker.send(message)
        self.after_emit("send", message=message)

    def before_emit(self, signal, *args, **kwargs):
        signal = "before_" + signal
        self.emit(signal, *args, **kwargs)

    def after_emit(self, signal, *args, **kwargs):
        signal = "after_" + signal
        self.emit(signal, *args, **kwargs)

    def emit(self, signal, *args, **kwargs):
        for middleware in self.middlewares:
            if not hasattr(middleware, signal):
                continue
            try:
                getattr(middleware, signal)(step=self, *args, **kwargs)
            except StopMiddleware as e:
                logger.debug(f"middleware<{middleware}> is stopped，reason: {e}")
                break

    @classmethod
    def is_shutdown(cls, group):
        # check all broker
        _consumers = cls._find_consumers(group)
        if not _consumers:
            return True
        return all(broker._shutdown for broker in _consumers)


def decorator_func_proxy(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    return wrapper


class SyncOneStep(BaseOneStep):

    def __call__(self, *args, **kwargs):
        """同步执行原函数"""

        result = self.fn(*args, **kwargs)

        if isgenerator(result):
            for item in result:
                self.send(item)
        else:
            self.send(result)
        return result


class AsyncOneStep(BaseOneStep):

    async def __call__(self, *args, **kwargs):
        """"异步执行原函数"""

        if iscoroutinefunction(self.fn):
            result = await self.fn(*args, **kwargs)
        else:
            result = self.fn(*args, **kwargs)

        if isasyncgen(result):
            async for item in result:
                self.send(item)
        else:
            self.send(result)
        return result


class step:

    def __init__(self,
                 *,
                 group: str = "OneStep",
                 name: str = None,
                 from_broker: Union[BaseBroker, List[BaseBroker], None] = None,
                 to_broker: Union[BaseBroker, List[BaseBroker], None] = None,
                 workers: Optional[int] = None,
                 worker_class: Optional[Type[BaseWorker]] = None,
                 middlewares: Optional[List[Any]] = None,
                 retry: Union[Callable, object] = TimesRetry(),
                 error_callback: Optional[Union[Callable, object]] = None):
        self.params = {
            "group": group,
            "name": name,
            "from_broker": from_broker,
            "to_broker": to_broker,
            "workers": workers,
            "worker_class": worker_class,
            "middlewares": middlewares,
            "retry": retry,
            "error_callback": error_callback
        }

    def __call__(self, func, *_args, **_kwargs):
        if iscoroutinefunction(func) or isasyncgenfunction(func):
            os = AsyncOneStep(fn=func, **self.params)
        else:
            os = SyncOneStep(fn=func, **self.params)

        return os.wraps(func)

    @staticmethod
    def start(group=None, block=None, print_jobs=False):
        BaseOneStep.start(group=group, print_jobs=print_jobs)
        started.send()
        if block:
            import time
            while not BaseOneStep.is_shutdown(group):
                time.sleep(1)

    @staticmethod
    def shutdown(group=None):
        BaseOneStep.shutdown(group=group)
        stopped.send()

    @staticmethod
    def set_debugging():

        if not BaseOneStep.state.debug:
            onestep_logger = logging.getLogger("onestep")
            handler = logging.StreamHandler()
            handler.setFormatter(logging.Formatter(
                "[%(levelname)s] %(asctime)s %(name)s:%(message)s"))
            onestep_logger.addHandler(handler)
            onestep_logger.setLevel(logging.DEBUG)
            BaseOneStep.state.debug = True



from abc import ABC, abstractmethod
from enum import Enum
from typing import Optional, Tuple, Union, Type

from .exception import RetryInLocal, RetryInQueue
from .message import Message


class RetryStatus(Enum):
    CONTINUE = 1  # 继续（执行重试）
    END_WITH_CALLBACK = 2  # 结束（执行回调）
    END_IGNORE_CALLBACK = 3  # 结束（忽略回调）


class BaseRetry(ABC):

    @abstractmethod
    def __call__(self, *args, **kwargs) -> Optional[RetryStatus]:
        pass


class NeverRetry(BaseRetry):

    def __call__(self, message) -> Optional[RetryStatus]:
        return RetryStatus.END_WITH_CALLBACK


class AlwaysRetry(BaseRetry):

    def __call__(self, message) -> Optional[RetryStatus]:
        return RetryStatus.CONTINUE


class TimesRetry(BaseRetry):

    def __init__(self, times: int = 3):
        self.times = times

    def __call__(self, message) -> Optional[RetryStatus]:
        if message.failure_count < self.times:
            return RetryStatus.CONTINUE
        return RetryStatus.END_WITH_CALLBACK


class RetryIfException(BaseRetry):

    def __init__(self, exceptions: Optional[Tuple[Union[Exception, Type]]] = Exception):
        self.exceptions = exceptions

    def __call__(self, message) -> Optional[RetryStatus]:
        if isinstance(message.exception.exc_value, self.exceptions):
            return RetryStatus.CONTINUE
        return RetryStatus.END_WITH_CALLBACK


class AdvancedRetry(TimesRetry):
    """高级重试策略
    
    1. 本地重试：如果异常是 RetryInLocal 或 指定异常，且重试次数未达到上限，则本地重试，不回调
    2. 队列重试：如果异常是 RetryInQueue，且重试次数未达到上限，则入队重试，不回调
    3. 其他异常：如果异常不是 RetryInLocal 或 RetryInQueue 或 指定异常，则不重试，回调
    注：待重试的异常若继承自 RetryException，则可单独指定重试次数，否则默认为 3 次
    """

    def __init__(self, times: int = 3, exceptions: Optional[Tuple[Union[Exception, Type]]] = None):
        super().__init__(times=times)
        self.exceptions = (RetryInLocal, RetryInQueue) + (exceptions or ())

    def __call__(self, message: Message) -> Optional[RetryStatus]:
        if isinstance(message.exception.exc_value, self.exceptions):
            max_retry_times = getattr(message.exception.exc_value, "times", None) or self.times
            if message.failure_count < max_retry_times:
                if isinstance(message.exception.exc_value, RetryInQueue):
                    return RetryStatus.END_IGNORE_CALLBACK
                return RetryStatus.CONTINUE
            else:
                return RetryStatus.END_WITH_CALLBACK
        else:
            return RetryStatus.END_WITH_CALLBACK


# error callback
class BaseErrorCallback(ABC):

    @abstractmethod
    def __call__(self, *args, **kwargs):
        pass


class NackErrorCallBack(BaseErrorCallback):

    def __call__(self, message):
        message.reject()



# -*- coding: utf-8 -*-
import abc
import logging
from queue import Queue, Empty
from typing import Any, Optional, List, Callable

from onestep.middleware import BaseMiddleware
from onestep.exception import StopMiddleware
from onestep.message import Message

logger = logging.getLogger(__name__)


class BaseBroker:
    message_cls = Message

    def __init__(self,
                 name: Optional[str] = None,
                 queue: Optional[Queue] = None,
                 middlewares: Optional[List[BaseMiddleware]] = None,
                 once: bool = False,
                 cancel_consume: Optional[Callable] = None):
        """
        @param name: broker name
        @param queue: broker queue name
        @param middlewares: broker middlewares
        @param once: just run once
            if once is True, when broker receive a message, it will shutdown
        @param cancel_consume: cancel consume
            cancel_consume(message) -> bool
            if cancel_consume return True, broker will shutdown
        """
        self.queue = queue
        self.name = name or "broker"
        self.middlewares = []
        self.once = once
        self.cancel_consume = cancel_consume

        if middlewares:
            for middleware in middlewares:
                self.add_middleware(middleware)

    def add_middleware(self, middleware: BaseMiddleware):
        self.middlewares.append(middleware)

    def send(self, message):
        """对消息进行预处理，然后再发送"""
        if not isinstance(message, Message):
            message = self.message_cls(body=message)
        # TODO: 对消息发送进行N次重试，确保消息发送成功。
        return self.publish(message.to_json())

    @abc.abstractmethod
    def publish(self, message: Any):
        """
        将消息原样发布到 broker 中。如果当前Broker是Job的to_broker, 则必须实现此方法
        """
        raise NotImplementedError('Please implement in subclasses.')

    @abc.abstractmethod
    def consume(self, *args, **kwargs):
        """
        消费消息。如果当前Broker是Job的from_broker, 则必须实现此方法
        """
        raise NotImplementedError('Please implement in subclasses.')

    @abc.abstractmethod
    def confirm(self, message: Message):
        """确认消息"""
        raise NotImplementedError('Please implement in subclasses.')

    @abc.abstractmethod
    def reject(self, message: Message):
        """拒绝消息"""
        raise NotImplementedError('Please implement in subclasses.')

    @abc.abstractmethod
    def requeue(self, message: Message, is_source=False):
        """
        重发消息：先拒绝 再 重入
        is_source = False 重入使用消息的当前状态
        is_source = True 重入使用消息的初始状态
        """
        raise NotImplementedError('Please implement in subclasses.')

    def before_emit(self, signal, *args, **kwargs):
        signal = "before_" + signal
        self._emit(signal, *args, **kwargs)

    def after_emit(self, signal, *args, **kwargs):
        signal = "after_" + signal
        self._emit(signal, *args, **kwargs)

    def _emit(self, signal, *args, **kwargs):
        for middleware in self.middlewares:
            if not hasattr(middleware, signal):
                continue
            try:
                getattr(middleware, signal)(self, *args, **kwargs)
            except StopMiddleware:
                break

    def shutdown(self):
        """关闭Broker"""

    def __repr__(self):
        return f"<{self.__class__.__name__} {self.name}>"

    def __str__(self):
        return self.name


class BaseConsumer:

    def __init__(self, broker: BaseBroker, *args, **kwargs):
        self.queue = broker.queue
        self.message_cls = broker.message_cls or Message
        self.timeout = kwargs.pop("timeout", 1000)

    def __next__(self):
        try:
            data = self.queue.get(timeout=self.timeout / 1000)
            return self.message_cls.from_broker(broker_message=data)
        except Empty:
            return None

    def __iter__(self):
        return self



from .stream import RedisStreamBroker, RedisStreamConsumer
from .pubsub import RedisPubSubBroker, RedisPubSubConsumer

__all__ = [
    "RedisStreamBroker",
    "RedisStreamConsumer",
    "RedisPubSubBroker",
    "RedisPubSubConsumer"
]



from .onestep import step
from .retry import (
    BaseRetry, BaseErrorCallback, NackErrorCallBack,
    NeverRetry, AlwaysRetry, TimesRetry, RetryIfException, AdvancedRetry
)
from .broker import (
    BaseBroker, BaseConsumer,
    MemoryBroker, RabbitMQBroker, WebHookBroker, CronBroker, RedisStreamBroker, RedisPubSubBroker
)
from .middleware import (
    BaseMiddleware, BaseConfigMiddleware,
    NacosPublishConfigMiddleware, NacosConsumeConfigMiddleware,
    RedisPublishConfigMiddleware, RedisConsumeConfigMiddleware,
)
from .exception import (
    StopMiddleware, DropMessage,
    RetryException, RetryInQueue, RetryInLocal
)

__all__ = [
    'step',

    # broker
    'BaseBroker',
    'BaseConsumer',
    'BaseLocalBroker',
    'BaseLocalConsumer',
    'MemoryBroker',
    'RabbitMQBroker',
    'WebHookBroker',
    'CronBroker',
    'RedisStreamBroker',
    'RedisPubSubBroker',

    # retry
    'BaseRetry',
    'NeverRetry',
    'AlwaysRetry',
    'TimesRetry',
    'RetryIfException',
    'AdvancedRetry',
    # error callback
    'BaseErrorCallback',
    'NackErrorCallBack',

    # middleware
    'BaseMiddleware',
    'BaseConfigMiddleware',
    'NacosPublishConfigMiddleware',
    'NacosConsumeConfigMiddleware',
    'RedisPublishConfigMiddleware',
    'RedisConsumeConfigMiddleware',

    # exception
    'StopMiddleware',
    'DropMessage',
    'RetryException',
    'RetryInQueue',
    'RetryInLocal',

    '__version__'
]

__version__ = '0.3.81'



import json
import threading
from queue import Queue
from typing import Optional, Dict, Any

import amqpstorm

from .base import BaseBroker, BaseConsumer
from use_rabbitmq import useRabbitMQ as RabbitMQStore
from ..message import Message


class _RabbitMQMessage(Message):

    @classmethod
    def from_broker(cls, broker_message: amqpstorm.Message):
        try:
            message = json.loads(broker_message.body)
        except json.JSONDecodeError:
            message = {"body": broker_message.body}
        if not isinstance(message, dict):
            message = {"body": message}
        if "body" not in message:
            # 来自 外部的消息 可能没有 body, 故直接认为都是 message.body
            message = {"body": message}

        return cls(body=message.get("body"), extra=message.get("extra"), message=broker_message)


class RabbitMQBroker(BaseBroker):
    message_cls = _RabbitMQMessage

    def __init__(self, queue_name, params: Optional[Dict] = None, prefetch: Optional[int] = 1, auto_create=True, *args,
                 **kwargs):
        super().__init__(*args, **kwargs)
        self.queue_name = queue_name
        self.queue = Queue()
        params = params or {}
        self.client = RabbitMQStore(**params)
        if auto_create:
            self.client.declare_queue(self.queue_name)
        self.prefetch = prefetch
        self.threads = []

    def _consume(self, *args, **kwargs):
        def callback(message: amqpstorm.Message):
            self.queue.put(message)

        prefetch = kwargs.pop("prefetch", self.prefetch)
        self.client.start_consuming(queue_name=self.queue_name, callback=callback, prefetch=prefetch, **kwargs)

    def consume(self, *args, **kwargs):
        daemon = kwargs.pop('daemon', True)
        thread = threading.Thread(target=self._consume, *args, **kwargs)
        thread.daemon = daemon
        thread.start()
        self.threads.append(thread)
        return RabbitMQConsumer(self)

    def publish(self, message: Any):
        self.client.send(self.queue_name, message)

    def confirm(self, message: Message):
        """确认消息"""
        message.message.ack()

    def reject(self, message: Message):
        """拒绝消息"""
        message.message.reject(requeue=False)

    def requeue(self, message: Message, is_source=False):
        """
        重发消息：先拒绝 再 重入
        
        :param message: 消息
        :param is_source: 是否是原始消息，True: 使用原始消息重入当前队列，False: 使用消息的最新数据重入当前队列
        """
        if is_source:
            message.message.reject(requeue=True)
        else:
            message.message.reject(requeue=False)
            self.send(message)

    def shutdown(self):
        self.client.shutdown()
        for thread in self.threads:
            thread.join()


class RabbitMQConsumer(BaseConsumer):
    ...



import functools
import logging


def catch_error(return_val=None):
    """
    捕获异常装饰器
    """

    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except Exception as e:
                logging.debug(e)
                return return_val

        return wrapper

    return decorator


"""
使用CRON表达式触发任务执行
"""
import logging
import threading
from datetime import datetime
from typing import Any

from croniter import croniter

from .memory import MemoryBroker, MemoryConsumer

logger = logging.getLogger(__name__)


class CronBroker(MemoryBroker):
    _thread = None

    def __init__(self, cron, name=None, middlewares=None, body: Any = None, *args, **kwargs):
        super().__init__(name=name, middlewares=middlewares, *args, **kwargs)
        self.cron = cron
        self.itr = croniter(cron, datetime.now())
        self.next_fire_time = self.itr.get_next(datetime)
        self.body = body

    def _scheduler(self):
        if self.next_fire_time <= datetime.now():
            self.next_fire_time = self.itr.get_next(datetime)
            self.publish(self.body)

        self._thread = threading.Timer(interval=1, function=self._scheduler)
        self._thread.start()

    def consume(self, *args, **kwargs):
        self._scheduler()
        return CronConsumer(self, *args, **kwargs)

    def shutdown(self):
        self._thread.cancel()


class CronConsumer(MemoryConsumer):
    ...
